/* 
 *  GGI/DM default drawmap allocation
 *
 *  Copyright (c) 1999 Nicolai Haehnle - Locinga@pfaffenwinkel.de
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define DEBUG		1

#include <stdlib.h>
#include <string.h>

#include <ggi/internal/ggi-dl.h>
#include <ggi/internal/dm_int.h>
#include "default.h"


int merge_rects(dmrect **list, dmrect *newrc)
{
	dmrect *rc;
	
	GGIDPRINT("merge_rects: pos(%i/%i), size(%i/%i)\n",
		newrc->pos.x, newrc->pos.y, newrc->size.x, newrc->size.y);
	
	if (!newrc->size.x || !newrc->size.y) {
		free(newrc);
		return 0;
	}
	
	/* special case: list is empty */
	if (!*list) {
		*list = newrc->prev = newrc->next = newrc;
		return 0;
	}
	
	/* check if we can merge some rectangles */
	rc = *list;
	do {
		if (rc->size.x == newrc->size.x && rc->pos.x == newrc->pos.x) {
			GGIDPRINT("Checking (%i/%i) (%i/%i)\n",
				rc->pos.x, rc->pos.y, rc->size.x, rc->size.y);
				
			if (rc->pos.y + rc->size.y == newrc->pos.y) {
				rc->size.y += newrc->size.y;
				free(newrc);
				GGIDPRINT("merged to pos.y=%i/size.y=%i\n",
					rc->pos.y, rc->size.y);

			/* Now take this rect out of the list and recurse
			 * to look whether we can merge again */
				if (rc != rc->next) {
					REMOVE_LIST(list, rc);
					GGIDPRINT("recurse\n");
					merge_rects(list, rc);
				}
				return 0;
			}

			if (newrc->pos.y + newrc->size.y == rc->pos.y) {
				newrc->size.y += rc->size.y;
				
				GGIDPRINT("merged to pos.y=%i/size.y=%i\n",
					rc->pos.y, rc->size.y);
					
				/* fix the list */
				REMOVE_LIST(list, rc);
				GGIDPRINT("recurse\n");
				merge_rects(list, newrc);
				free(rc);
				return 0;
			}
		}

	/* Now the same, but for horizontal merging */
		if (rc->size.y == newrc->size.y && rc->pos.y == newrc->pos.y) {
			GGIDPRINT("Checking (%i/%i) (%i/%i)\n",
				rc->pos.x, rc->pos.y, rc->size.x, rc->size.y);
				
			if (rc->pos.x + rc->size.x == newrc->pos.x) {
				rc->size.x += newrc->size.x;
				free(newrc);
				GGIDPRINT("merged to pos.x=%i/size.x=%i\n",
					rc->pos.x, rc->size.x);

			/* Now take this rect out of the list and recurse
			 * to look whether we can merge again */
				if (rc != rc->next) {
					REMOVE_LIST(list, rc);
					GGIDPRINT("recurse\n");
					merge_rects(list, rc);
				}
				return 0;
			}

			if (newrc->pos.x + newrc->size.x == rc->pos.x) {
				newrc->size.x += rc->size.x;
				
				GGIDPRINT("merged to pos.x=%i/size.x=%i\n",
					rc->pos.x, rc->size.x);
					
				/* fix the list */
				REMOVE_LIST(list, rc);
				GGIDPRINT("recurse\n");
				merge_rects(list, newrc);
				free(rc);
				return 0;
			}
		}

		rc = rc->next;
	} while(rc != *list);
	
	/* no way to merge it - simply put it into the list */
	INSERT_LIST(list, newrc);
	
	return 0;
}

static int remove_drawmap(ggidm_drawmap dm)
{
	dmext *hook;
	dmrect *rc;
	dmvisual *dv;
	
	GGIDPRINT("default-dm: removing drawmap (%p)\n", dm);
	
	if (!dm || !(dm->flags & GGIDM_ACQUIRED) || !dm->parent ||
	    dm->flags & GGIDM_VISIBLE)
		return -1;
	
	dv = dm->parent;
	dm->flags &= ~GGIDM_ACQUIRED; /* No longer acquired */
	
	/* remove from dmvis list */
	REMOVE_LIST(&dv->drawmaps, dm);
	
	/* put into dmext list */
	hook = dv->parent;
	INSERT_LIST(&hook->drawmaps, dm);
	
	if (!dv->drawmaps && !(dv->flags & GGIDM_VISIBLE)) {
		/* visual is empty now */
		free_visual(hook, dv);
		return 0;
	}
	
	/* update freerects list */
	rc = (dmrect *)malloc(sizeof(dmrect));
	if (!rc) {
		GGIDPRINT("default-dm: Out of memory\n");
		return -1;
	}
	memset(rc, 0, sizeof(*rc));

	rc->pos.x = dm->pos.x;
	rc->pos.y = dm->pos.y;
	rc->size.x = dm->size.x;
	rc->size.y = dm->size.y;
	merge_rects(&dv->freerects, rc);

	return 0;
}



/* This tries to put the drawmap dm into the visual with the freerects
 * list "list".
 * Following return values are possible:
 *  0 success (only position is altered)
 * -1 no space left
 *  1 space free at the border -> resize by (ex/ey) and retry
 *
 * Big Fat TODO:
 *  - Implement search for rectangles at the edges
 *  - Maybe optimize search for mergeable rectangles
 *  - Test!!!
 */

#define SPLIT_RIGHT(list,rc,width) \
do {									\
	dmrect *__rc;							\
	if (!(width)) break;						\
	(rc)->size.x -= width;						\
	__rc = (dmrect *)malloc(sizeof(dmrect));			\
	__rc->pos.y = (rc)->pos.y; 					\
	__rc->pos.x = (rc)->pos.x + (rc)->size.x;			\
	__rc->size.y = (rc)->size.y; __rc->size.x = width;		\
	merge_rects(list,(rc));						\
} while(0)
 
#define SPLIT_BOTTOM(list,rc,height) \
do {									\
	dmrect *__rc;							\
	if (!(height)) break;						\
	(rc)->size.y -= height;						\
	__rc = (dmrect *)malloc(sizeof(dmrect));			\
	__rc->pos.y = (rc)->pos.y+(rc)->size.y; 			\
	__rc->pos.x = (rc)->pos.x;					\
	__rc->size.y = height; __rc->size.x = (rc)->size.x;		\
	merge_rects(list,(rc));						\
} while(0)
 
#define SPLIT_LEFT(list,rc,width) \
do {									\
	dmrect *__rc;							\
	if (!(width)) break;						\
	(rc)->size.x -= width; (rc)->pos.x += width;			\
	__rc = (dmrect *)malloc(sizeof(dmrect));			\
	__rc->pos.y = (rc)->pos.y; __rc->pos.x = (rc)->pos.x;		\
	__rc->size.y = (rc)->size.y; __rc->size.x = width;		\
	merge_rects(list,(rc));						\
} while(0)
 
#define SPLIT_TOP(list,rc,height) \
do {									\
	dmrect *__rc;							\
	if (!(height)) break;						\
	(rc)->size.y -= height; (rc)->pos.y += height;			\
	__rc = (dmrect *)malloc(sizeof(dmrect));			\
	__rc->pos.y = (rc)->pos.y; __rc->pos.x = (rc)->pos.x;		\
	__rc->size.y = (height); __rc->size.x = (rc)->size.x;		\
	merge_rects(list,(rc));						\
} while(0)

/*
 * Rectangle base is wide enough for dm, but too low. Check if we can
 * merge something
 *
 * It does the following steps:
 *  1. Check if mixing is really possible
 *  2. Split the appropriate parts:
 *   a) left and right part of top
 *   b) left and right part of down
 *   c) bottom part of down
 *  3. Fill in dm structure
 *  4. Remove the original rectangles
 */
#define CM_VERTICALLY(list, top, down, dm) \
do {									\
	int cxl, cxr;							\
	cxl = top->pos.x; if (cxl < down->pos.x) cxl = down->pos.x;	\
	cxr = top->pos.x+top->size.x;					\
	if (cxr > down->pos.x+down->size.x) 				\
		cxr = down->pos.x+down->size.x;				\
	if ((cxr-cxl) < dm->size.x)					\
		break;							\
	cxr = cxl+dm->size.x;						\
									\
	SPLIT_LEFT(list, top, cxl-top->pos.x);				\
	SPLIT_RIGHT(list, top, top->pos.x+top->size.x-cxr);		\
	SPLIT_LEFT(list, down, cxl-down->pos.x);			\
	SPLIT_RIGHT(list, down, down->pos.x+down->size.x-cxr);		\
	SPLIT_BOTTOM(list, down, top->size.y+down->size.y-dm->size.y);	\
									\
	dm->pos.x = top->pos.x; dm->pos.y = top->pos.y;			\
									\
	REMOVE_LIST(list, top);						\
	REMOVE_LIST(list, down);					\
	free(top); free(down);						\
									\
	return 0;							\
} while(0)


static int cut_merge_vertical(dmrect **list, ggidm_drawmap dm, dmrect *base)
{
	dmrect *rc;
	
	GGIDPRINT("called cut_merge_vertical(%p, (%i/%i), (%i/%i|%i/%i)\n",
		list, dm->size.x, dm->size.y, base->pos.x, base->pos.y,
		base->size.x, base->size.y);
	
	rc = base->next;
	while(rc != base) {
		if ((rc->size.y + base->size.y) >= dm->size.y) {
			if (rc->pos.y+rc->size.y == base->pos.y) {
				CM_VERTICALLY(list, rc, base, dm);
			}
			if (base->pos.y+base->size.y == rc->pos.y) {
				CM_VERTICALLY(list, base, rc, dm);
			}
		}
	
		rc = rc->next;
	}
	
	GGIDPRINT("No success\n");
	
	return -1;
}

#define CM_HORIZONTALLY(list, left, right, dm) \
do {									\
	int cyt, cyd;							\
	cyt = left->size.y; if (cyt < right->pos.y) cyt = right->pos.y;	\
	cyd = left->size.y+left->size.y;				\
	if (cyd > right->pos.y+right->size.y) 				\
		cyd = right->pos.y+right->size.y;			\
	if ((cyd-cyt) < dm->size.y)					\
		break;							\
	cyd = cyt + dm->size.y;						\
									\
	SPLIT_TOP(list, left, cyt-left->pos.y);				\
	SPLIT_BOTTOM(list, left, left->pos.y+left->size.y-cyd);		\
	SPLIT_TOP(list, right, cyt-right->pos.y);			\
	SPLIT_BOTTOM(list, right, right->pos.y+right->size.y-cyd);	\
	SPLIT_RIGHT(list,right,left->size.x+right->size.x-dm->size.x);	\
									\
	dm->pos.x = left->pos.x; dm->pos.y = left->pos.y;		\
									\
	REMOVE_LIST(list, left);					\
	REMOVE_LIST(list, right);					\
	free(left); free(right);					\
									\
	return 0;							\
} while(0)


static int cut_merge_horizontal(dmrect **list, ggidm_drawmap dm, dmrect *base)
{
	dmrect *rc;
	
	GGIDPRINT("called cut_merge_horizontal(%p, (%i/%i), (%i/%i|%i/%i)\n",
		list, dm->size.x, dm->size.y, base->pos.x, base->pos.y,
		base->size.x, base->size.y);
	
	rc = base->next;
	while(rc != base) {
		if ((rc->size.x + base->size.x) >= dm->size.x) {
			if (rc->pos.x+rc->size.x == base->pos.x) {
				CM_HORIZONTALLY(list, rc, base, dm);
			}
			if (base->pos.x+base->size.x == rc->pos.x) {
				CM_HORIZONTALLY(list, base, rc, dm);
			}
		}
	
		rc = rc->next;
	}
	
	GGIDPRINT("No success\n");
	
	return -1;
}

static int cut_rects(dmrect **list, ggidm_drawmap dm, int *ex, int *ey)
{
	dmrect *rc;
	
	GGIDPRINT("cut_rects: size(%i/%i)\n", dm->size.x, dm->size.y);
	
	/* No free rectangle => have to resize visual */
	if (!*list)
		return -1;
	
	/* First run: check for rectangles that are big enough */
	rc = *list;
	do {
		if (rc->size.x >= dm->size.x && rc->size.y >= dm->size.y) {
			SPLIT_RIGHT(list, rc, rc->size.x-dm->size.x);
			SPLIT_BOTTOM(list, rc, rc->size.y-dm->size.y);
			
			REMOVE_LIST(list, rc);
			dm->pos.x = rc->pos.x;
			dm->pos.y = rc->pos.y;
			free(rc);
			return 0;
		}
	
		rc = rc->next;
	} while(rc != *list);
	
	/* Second run: check if merging several rectangles helps */
	GGIDPRINT("First run failed\n");
	
	rc = *list;
	do {
		if (rc->size.x >= dm->size.x) {
			if (!cut_merge_vertical(list, dm, rc))
				return 0;
		}
		if (rc->size.y >= dm->size.y) {
			if (!cut_merge_horizontal(list, dm, rc))
				return 0;
		}
	
		rc = rc->next;
	} while(rc != *list);
	
	/* Third run: look for rectangles at the edges */
	GGIDPRINT("Second run failed\n");
	
	return -1;
}

/*
 * *Tries* to insert drawmap dm into visual dv. It only sets
 * dm->pos and dm->vis, dm->parent and updates the freerects of the visual.
 * The linked lists aren't affected
 */
int insert_drawmap(dmvisual *dv, ggidm_drawmap dm)
{
	int ex, ey;
	int result;

	GGIDPRINT("inserting drawmap %p: size(%i/%i)\n", 
		dm, dm->size.x, dm->size.y);

/* Look for freerects that are big enough */
 retry:
	result = cut_rects(&dv->freerects, dm, &ex, &ey);
	switch(result) {
	 case -1:
	  	/* Bad luck. Trying to resize the visual. But first, check
	  	 * if we're allowed to do this */
	  	if (dv->flags & GGIDM_ACQUIRED)
	  		return -1;	/* No space left */
	  	
	  	if (dv->size.x < dm->size.x) {
	  		GGIDPRINT("Resize horizontally\n");
	  		if (!expand_visual(dv, dm->size.x-dv->size.x, 0))
	  			goto retry;
	  	}
	  	if (dv->size.y < dm->size.y) {
	  		GGIDPRINT("Resize vertically\n");
	  		if (!expand_visual(dv, 0, dm->size.y-dv->size.y))
	  			goto retry;
	  	}
	  	GGIDPRINT("Resize visual vertically\n");
	 	if (!expand_visual(dv, 0, dm->size.y))
	 		goto retry;	/* resize worked */
	 	GGIDPRINT("Resize visual horizontally\n");
	 	if (!expand_visual(dv, dm->size.x, 0))
	 		goto retry;
	 	GGIDPRINT("Resizing failed\n");
	 	
	 	return -1; /* Couldn't resize - giving up */
	 case 1:
	 	/* No space left, but there's space left at the edge of
	 	 * the visual. Resize it a little bit, if possible */
	 	if (dv->flags & GGIDM_ACQUIRED)
	 		return -1;
	 	
	 	GGIDPRINT("Resize visual by (%i/%i) to (%i/%i)\n",
	 		ex, ey, dv->size.x+ex, dv->size.y+ey);
	 	if (!expand_visual(dv, ex, ey))
	 		goto retry;
	 	
	 	return -1;
	 default:	/* Success */
	}

	dm->flags |= GGIDM_INVALID | GGIDM_ACQUIRED;
	dm->vis = dv->vis;
	dm->parent = dv;
	return 0;
}


/*
 * Creates a drawmap of the given size and storage type (flags)
 * It doesn't get acquired. Acquire it by either setting a mode
 * or using ggidmAcquire() manually
 */
ggidm_drawmap GGIDMadddrawmap(ggi_visual_t vis, int sizex, int sizey,
			      int flags)
{
	dmext *hook = LIBGGI_DMEXT(vis);
	ggidm_drawmap dm = (ggidm_drawmap)malloc(sizeof(*dm));
	memset(dm, 0, sizeof(*dm));

	GGIDPRINT("called GGIDMadddrawmap(%p, %i, %i, %08X)\n", 
		vis, sizex, sizey, flags);
	
	if (sizex < 0 || sizey < 0) {
		free(dm);
		return 0;
	}
	
	dm->hook = hook;
	
	dm->flags = flags & GGIDM_STORE_MASK;
	dm->size.x = sizex;
	dm->size.y = sizey;
	
	INSERT_LIST(&hook->drawmaps, dm);
	
	GGIDPRINT("Done: dm=%p, dm->next=%p\n", dm, dm->next);
	
	return dm;
}

/*
 * Close the given drawmap. You don't have to unacquire it first.
 * You can't close visible drawmaps. These are automatically destroyed
 * on setmode.
 * Of course, the pointer to dm gets invalid.
 */
int GGIDMclosedrawmap(ggidm_drawmap dm)
{
	int error;
	dmext *hook;

	GGIDPRINT("called GGIDMclosedrawmap(%p)\n", dm);
	
	hook = dm->parent->parent;
	
	if (dm->flags & GGIDM_VISIBLE) {
		GGIDPRINT("Can't close visible drawmap\n");
		return GGIDM_ERR_INVALID;
	}
	if (dm->flags & GGIDM_ACQUIRED) {
		GGIDPRINT("calling __GGIDMacquire(%p, 0)\n", dm);
		error = __GGIDMacquire(hook, dm, 0);
		if (error) {
			GGIDPRINT("Failed\n");
			return error;
		}
	}
	
	/*
	 * The drawmap dm *must* be in dmext->drawmaps now. We rely
	 * on __GGIDMacquire() here.
	 */
	GGIDPRINT("Removing from list and destroying it\n");
	
	REMOVE_LIST(&hook->drawmaps, dm);
	free(dm);
	
	GGIDPRINT("Done\n");
	return 0;
}


/*
 * GGIDMacquire:
 *
 * (Un-)Acquires drawmap dm (depending on flag acquire)
 * When unacquiring, the GGIDM_ACQUIRED-flag is cleared. All following
 *  drawing operations will result in a GGIDM_ERR_NOTACQUIRED-error.
 * When acquiring, the GGIDM_INVALID-flag is set. All following drawing
 *  operations with this drawmap as source will return GGIDM_ERR_INVALIDATED.
 *  All drawing operations with this drawmaps as destination will clear
 *  the GGIDM_INVALID-flag.
 */
 
#define CHECK_STORAGE(dmflags, dvflags) \
 (((dmflags) == GGIDM_STORE_DONTWORRY) ||				\
  (((dmflags) == GGIDM_STORE_VIDEOMEMORY) && (dvflags & GGIDM_VISIBLE)) || \
  (((dmflags) == GGIDM_STORE_SYSTEMMEMORY) && 				\
  	!(dvflags & GGIDM_VISIBLE)))
static int do_acquire(dmext *hook, ggidm_drawmap dm)
{
	dmvisual *dv;
	
	GGIDPRINT("called acquire(flags=%08X)\n", dm->flags);
	
	/* Remove from unacquired list */
	REMOVE_LIST(&hook->drawmaps, dm);
	
	dv = hook->dmvis;
	do {
		GGIDPRINT("Trying visual %p (flags=%08X)\n", dv, dv->flags);
		
		if (!CHECK_STORAGE(dm->flags & GGIDM_STORE_MASK, dv->flags)) {
			GGIDPRINT("Incompatible storage types\n");
		} else {
			if (insert_drawmap(dv, dm)) {
				GGIDPRINT("Out of space\n");
			} else {
				GGIDPRINT("Got it\n");
				INSERT_LIST(&dv->drawmaps, dm);
				return 0;
			}
		}
	
		dv = dv->next;
	} while(dv != hook->dmvis);
	
	GGIDPRINT("Couldn't find enough space. Trying new visual\n");
	if (!CHECK_STORAGE(dm->flags & GGIDM_STORE_MASK, 0)) {
		return GGIDM_ERR_OUTOFVIDEOMEMORY;
	}
	
	if (!(dv = create_visual(hook)))
		return GGIDM_ERR_OUTOFVIDEOMEMORY;
	if (insert_drawmap(dv, dm)) {
		GGIDPRINT("Huh? Seems strange.. \n");
		return GGIDM_ERR_INTERNAL;
	}
	INSERT_LIST(&dv->drawmaps, dm);
	
	return 0;
}


int __GGIDMacquire(dmext *hook, ggidm_drawmap dm, int acquire)
{
	int error;
	
	GGIDPRINT("called __GGIDMacquire(%p, %i)\n", dm, acquire);
	
	if (dm->flags & GGIDM_VISIBLE) {
		GGIDPRINT("Can't acquire visible drawmap\n");
		return GGIDM_ERR_BADDRAWMAP;
	}
	
	if (acquire) {
		if (dm->flags & GGIDM_ACQUIRED) {
			GGIDPRINT("Drawmap already acquired\n");
			return 0;
		}
		error = do_acquire(hook, dm);
		if (error) {
			GGIDPRINT("Failed\n");
			return error;
		}
		dm->flags |= GGIDM_ACQUIRED | GGIDM_INVALID;
		return 0;
	}
	
	if (!(dm->flags & GGIDM_ACQUIRED)) {
		GGIDPRINT("Drawmap already unacquired\n");
		return 0;
	}
	if (remove_drawmap(dm)) {
		GGIDPRINT("Failed\n");
		return GGIDM_ERR_INTERNAL; /* shouldn't happen */
	}
	dm->flags &= ~GGIDM_ACQUIRED;
	
	GGIDPRINT("Done\n");
	return 0;
}

int GGIDMacquire(ggidm_drawmap dm, int acquire)
{
	dmext *hook = dm->hook;
	int error;
	dmvisual *dv;

	GGIDPRINT("GGIDMacquire called, passing on...\n");

	error = __GGIDMacquire(hook, dm, acquire);
	if (error)
		return error;
	
	if (acquire) {
		dv = dm->parent;
		if (!(dv->flags & GGIDM_ACQUIRED)) {
			GGIDPRINT("Setting mode on parent visual\n");
			if ((error = physically_set_mode(dv->parent, dv))) {
				return error;
			}
		}
	}
	
	GGIDPRINT("Done\n");
	return 0;
}

